/*****************************************************************************
 *                                                                           *
 *  Program:      Namegen                                                    *
 *                                                                           *
 *  Description:  Template-based, random name generator                      *
 *                                                                           *
 *  File:         namegen.c, which contains the following subroutines or     *
 *                functions:                                                 *
 *                  init_letters()                                           *
 *                  to_letter()                                              *
 *                  to_index()                                               *
 *                  load_template()                                          *
 *                  command_line()                                           *
 *                  main()                                                   *
 *                  max_syllable()                                           *
 *                  proper_word()                                            *
 *                  usage()                                                  *
 *                  print_structure() **DEBUG**                              *
 *                                                                           *
 *  Copyright 1991 by Mark F. Cook and Hewlett-Packard, Corvallis            *
 *  Workstation Operation.  All rights reserved.                             *
 *                                                                           *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define MAX_LEN	50  /* max. length of a name */
#define NUM_ALPHA	26
#define NUM_SPECIAL	 3
#define RANGE  	(NUM_ALPHA+NUM_SPECIAL+1)

#define TRUE		 1
#define FALSE		 0
#define END_OF_WORD	 0

#define RESET		 0
#define START		 1
#define TEST		 2

static char special[] = "-' ";
static char vowel[] = "aeiouy";
static char consonant[] = "bcdfghjklmnpqrstvwxz";

unsigned int first_letter[RANGE];
	/* occurance of 1st letters in a word */
unsigned int second_letter[RANGE][RANGE];
	/* occurance of 2nd letters in a word */
unsigned int third_letter[RANGE][RANGE][RANGE];
	/* occurance of 3rd-and-beyond letters in a word */
char *progname;

extern int random();  /* randomly generate a number range 1-n */
int num_names = 0, max_user_syllables = 99;
char template_file[80];
int debug = TRUE;

/*****************************************************************************
 * Initialize the 1, 2, and 3-dimensional character frequency arrays to 0.   *
 *****************************************************************************/

init_letters()
{
  register int i, j, k;

  for (i=0; i<RANGE; i++) {
    first_letter[i] = 0;
    for (j=0; j<RANGE; j++) {
      second_letter[i][j] = 0;
      for (k=0; k<RANGE; k++)
	third_letter[i][j][k] = 0;
     }
   }
}


/*****************************************************************************
 * Convert an index in the range 0-(NUM_ALPHA+NUM_SPECIAL) back to the       *
 * correct char.                                                             *
 *****************************************************************************/

char to_letter(n)
int n;
{
  if (n == 0)
    return ((char) n);  /* END-OF-WORD index; return unmodified */
  if ((n > 0) && (n < 27))
    return ((char) (n-1) + 'a');
  return(special[n-27]);
}


/*****************************************************************************
 * Convert a character (ALPHA or SPECIAL) to the corresponding int. index.   *
 *****************************************************************************/

int to_index(c)
char c;
{
  char ch, *c_ptr;

  if ((c_ptr = strchr(special, c)) != NULL)
    return (c_ptr-special + 27);
  ch = (char) tolower(c);
  if ((ch >= 'a') && (ch <= 'z'))
    return(ch - 'a' + 1);
  return(0);
}


/*****************************************************************************
 * Read all the names in the template file one at a time.  For each first    *
 * letter, increment the corresponding first_letter[] array element.  Do the *
 * same for each second letter in the second_letter[] array.  For all the    *
 * following letters, increment the appropriate element in the third_letter[]*
 * array, and then shift the indices up (3rd->2nd, 2nd->1st) for the next    *
 * letter.  Continue this until the end of word.  Repeat for each word until *
 * EOF.                                                                      *
 *****************************************************************************/

load_template(fname)
char *fname;
{
  FILE *fd;
  char line[128], *status;
  int done, i, c1, c2, c3, len;
/*DEBUG*/ int linecount;

  fd = fopen(fname, "r");
  if (fd == NULL) {
    fprintf(stderr, "%s: Cannot open file \"%s\"\n", progname, fname);
    exit(-1);
   }

/*DEBUG*/ linecount = 0;
  done = FALSE;
  while (!done) {
    c1 = 0;  c2 = 0;  c3 = 0;
    status = fgets(line, 81, fd);
    if (status == NULL) done++;
    len = strlen(line);
    if (len < 2) continue;  /* throw out 1-letter template words */
    for (i=0; i<len; i++) {
      if (i == 0) {
	c1 = to_index(line[i]);
	first_letter[c1]++;
       }
      else if (i == 1) {
	c2 = to_index(line[i]);
	second_letter[c1][c2]++;
       }
      else {
	c3 = to_index(line[i]);
	third_letter[c1][c2][c3]++;
        c1 = c2;
	c2 = c3;
       }
     } /* for */
    third_letter[c1][c2][0]++; /* increment the 0th [END-OF-WORD] element */
/*DEBUG*/ linecount++;
   } /* while */
/*DEBUG*/ if (debug)
	fprintf(stdout, "Loaded %d template names\n", linecount);
}


/***************************************************************************
 *                                                                         *
 * Routine:   command_line                                                 *
 *                                                                         *
 * Purpose:   Parse all the command line options.                          *
 *                                                                         *
 *X11***********************************************************************/

command_line(num, cmd)
int num;
char *cmd[];
{
  int i;
  char *arg;

  for (i=1; i<num; i++) {
      arg = cmd[i];
      if (arg[0] == '-') {		/* process command line switches */
        switch (arg[1]) {
            case 'n':			/* get no. of names to generate */
                  if (++i >= num) usage();
                  num_names = atoi(cmd[i]);
                  continue;
            case 's':			/* get max. no. of syllables */
                  if (++i >= num) usage();
                  max_user_syllables = atoi(cmd[i]);
                  continue;
            default:
              usage();
          } /* switch */
       } /* if */
      else
	strcpy(template_file, cmd[i]);
   } /* for */
  return (TRUE);
}


main(int argc, char **argv) {
  int num_chars, i, j, c1, c2, c3;
  int current, total, start, stop, done;
  int end_of_word;
  char line[81];

  progname = argv[0];
 
  command_line(argc, argv);
  if ((num_names < 1) || (num_names > 50000)) {
    fprintf(stderr, "usage %s: -n <number> [-s <syllabes>] <template file>\n", progname);
    exit(-1);
   }

/*DEBUG*/
  if (debug) {
	fprintf(stdout, "Random test:");
	for (i=0; i<10; i++)
	  fprintf(stdout, " %d", random(20));
	fprintf(stdout, "\n");
	fprintf(stdout, "No. names = %d\n", num_names);
	fprintf(stdout, "Max. syllables = %d\n", max_user_syllables);
       }

  init_letters();
/*DEBUG*/
  if (debug)
	fprintf(stdout, "Arrays initialized\n");
  load_template(template_file);
/*DEBUG*/
  if (debug)
	fprintf(stdout, "Template file loaded\n");
  start = 0;
  for (i=0; i<RANGE; i++)
    start += first_letter[i];
/*--- execute this for_loop once for each word requested ---*/

  for (i=0; i<num_names; i++) {
    num_chars = 0;
    end_of_word = FALSE;
    max_syllable(0, RESET);

/*--- while the word is not greater than MAX_LEN letters long ---*/
/*--- and we haven't hit an end_of_word yet, grow the word.   ---*/

    while ((num_chars < MAX_LEN) && !end_of_word) {
/*--- generate the first letter of the word ---*/
      if (num_chars == 0) {
	stop = random(start);
	done = FALSE;
	current = 0;
	j = 0;
	while (!done) {
	  current += first_letter[j];
	  if (current >= stop) {
	    max_syllable(j, START);
	    sprintf(line, "%c", to_letter(j));
	    c1 = j;
	    done++;
	   } /* if */
	  j++;
	 } /* while(!done) */
       } /* if */
/*--- generate the second letter of the word ---*/
      else if (num_chars == 1) {
	total = 0;
	for (j=0; j<RANGE; j++)
	  total += second_letter[c1][j];
	stop = random(total);
	done = FALSE;
	current = 0;
	j = 0;
	while (!done) {
	  current += second_letter[c1][j];
	  if (current >= stop) {
	    max_syllable(j, TEST);
	    if (j == END_OF_WORD)
	      end_of_word++;
	    else
	      sprintf(&line[num_chars], "%c", to_letter(j));
	    c2 = j;
	    done++;
	   } /* if */
	  j++;
	 } /* while(!done) */
       } /* else if */
/*--- generate the third (and beyond) letter of the word ---*/
      else {
	total = 0;
	for (j=0; j<RANGE; j++)
	  total += third_letter[c1][c2][j];
	stop = random(total);
	done = FALSE;
	current = 0;
	j = 0;
	if (total == 0) {
	  done++;
	  end_of_word++;
	 }
	while (!done) {
	  current += third_letter[c1][c2][j];
	  if (current >= stop) {
	    if ((j == END_OF_WORD) || (max_syllable(j, TEST)))
	      end_of_word++;
	    else
	      sprintf(&line[num_chars], "%c", to_letter(j));
	    c3 = j;
	    done++;
	   } /* if */
	  j++;
	 } /* while(!done) */
        c1 = c2; c2 = c3;
       } /* else */
      num_chars++;
     } /* while */
    if (proper_word()) {
      for (j=0; j<num_chars-1; j++)
        fprintf(stdout, "%c", line[j]);
/*DEBUG*/ if (debug) {
		fprintf(stdout, " (%d) ", num_chars);
		print_structure();
	   }
      fprintf(stdout,"\n");
     }
    else i--;
   } /* for */
}


/*****************************************************************************
 *                                                                           *
 *  Function:  max_syllable                                                  *
 *                                                                           *
 *  Purpose:   To track the total number of 'syllables' so far in the word   *
 *             currently being generated, and to cut off the word if the     *
 *             total syllable count starts to exceed the max. syllable       *
 *             number specified by the user.                                 *
 *                                                                           *
 *             Syllables (for purposes of this routine) are as follows:      *
 *               C - a string of 1 or more consonants.                       *
 *               V - a string of 1 or more vowels.                           *
 *               S - a string of 1 or more special characters.               *
 *             The following are legal syllable combinations: CV, VC, CVC    *
 *             In cases such as CVC-CV, VC-CVC, and VC-VC, the trailing      *
 *             consonant group in the first syllable will also be used as    *
 *             the leading consonant group in the second syllable.           *
 *                                                                           *
 *             Special characters can occur after any alpha character and    *
 *             immediately indicate the end of the previous syllable.        *
 *                                                                           *
 *****************************************************************************/

#define BLANK		0
#define VOWEL		1
#define CONSONANT	2
#define SPECIAL		3

static int last_frag = BLANK, frag_count = 0;
static int frag_list[40], num_syllables;
static char syl_type[4] = { ' ', 'V', 'C', 'S' };

int max_syllable(n, flag)
int n, flag;
{
  char cur_ch;
  int i;

  cur_ch = to_letter(n);
  if (flag == RESET) {
    frag_count = 0;
    last_frag = BLANK;
    return (FALSE);
   }
  if (strchr(vowel, cur_ch)) {
    if (last_frag != VOWEL) {
      num_syllables = 0;
      for (i=0; i<frag_count; i++) {
        if (frag_list[i] == VOWEL)
	  num_syllables++;
       } /* for */
      if (num_syllables == max_user_syllables)
        return (TRUE);
      frag_list[frag_count++] = VOWEL;
      last_frag = VOWEL;
     } /* if(last_frag...) */
   } /* if(strchr...) */
  else if (strchr(consonant, cur_ch)) {
    if (last_frag != CONSONANT) {
      frag_list[frag_count++] = CONSONANT;
      last_frag = CONSONANT;
     }
   }
  else if (strchr(special, cur_ch)) {
    if (last_frag != SPECIAL) {
      frag_list[frag_count++] = SPECIAL;
      last_frag = SPECIAL;
     }
   }
  return (FALSE);
}


/*****************************************************************************
 *                                                                           *
 *  Function:  proper_word                                                   *
 *                                                                           *
 *  Purpose:   To determine whether the word just generated is a 'proper'    *
 *             word.  Proper words must not meet ANY of the following        *
 *             criteria:                                                     *
 *               1)  Word consists exclusively of consonants and special     *
 *                   characters.                                             *
 *               2)  Word ends in a special character.                       *
 *               3)  Word ends in a special character followed by only       *
 *                   consonants.                                             *
 *             If the word meets one or more of these criteria, return       *
 *             FALSE, otherwise return TRUE.                                 *
 *                                                                           *
 *****************************************************************************/

proper_word()
{
  int i, num_vowels, flag;

  flag = TRUE;

/*--- Test #1 ---*/
  num_vowels = 0;
  for (i=0; i<frag_count; i++) {
    if (frag_list[i] == VOWEL)
      num_vowels++;
   }
  if (!num_vowels) flag = FALSE;

/*--- Test #2 ---*/
  if (frag_list[frag_count-1] == SPECIAL) flag = FALSE;

/*--- Test #3 ---*/
  if ((frag_list[frag_count-2] == SPECIAL) &&
      (frag_list[frag_count-1] == CONSONANT)) flag = FALSE;

/*DEBUG*/ if (debug && !flag)
	fprintf(stdout, "*************word failed.\n");
  return(flag);
}


usage()
{
  fprintf(stderr,
	"Usage: %s [-n num_names] [-s max_syllables] [template_file]\n",
		progname);
  exit(-1);
}


print_structure()
{
  int i;

  fprintf(stdout, " - ");
  for (i=0; i<frag_count; i++)
    fprintf(stdout, "%c", syl_type[frag_list[i]]);
  fprintf(stdout, " -  %d", frag_count);
}


#ifdef BSD
#define MAX_RAND	2147483647
#else
#define MAX_RAND	32767
#endif

/* random(n) - a random integer generator with 20 bit resolution */

int random_first = 1;

random(n)	/* generate a number from 1 to n, n < 1.0EE6 */
int n;
{
	float fval;
	int val;

	if (random_first) {
		srand((unsigned int) time((long *)0));
		random_first=0;
	}
	fval = ((float) rand()) / MAX_RAND;
	val = fval * ((float) n);
	return(++val);
}
